<!--Edit this file for customizing the side bar of the front page-->
<a href="https://github.com/dpploy/cortix"><img src="https://cortix.org/logo.jpg" alt="Cortix" style="border:none;"></a>
<h3>About Cortix</h3>
<p>
Cortix is a Python library for network modeling and HPC simulation.
</p>
<p>
Cortix Group <br>
c/o <a href="https://www.uml.edu/innovation-hub">UMass Innovation Hub</a> <br>
110 Canal St., 3rd Floor <br>
Lower, MA 01852
</p>

<p>
<a href="https://www.uml.edu/Engineering/Chemical/faculty/de-Almeida-Valmor.aspx">Prof. V. F. de Almeida, Ph.D.</a></br>
Group Leader <br>
<a href="https://www.uml.edu/Engineering/Chemical">Dept. of Chemical Engineering (Nuclear Program)</a></br>
<a href="https://www.uml.edu/">UMass Lowell</a><br>
</p>

<!--{{ toctree() }}-->

<h3><a href="{{ pathto(master_doc) }}">{{ ('Table of Contents') }}</a></h3>
<ul>
<li><a href="{{ pathto('index') }}">Home</a></li>
<li><a href="{{ pathto('install') }}">Install</a></li>
<li><a href="{{ pathto('contents') }}">On-line Documentation</a></li>
<li><a href="https://dpploy.github.io/cortix/Cortix.pdf">PDF Documentation</a></li>
<li><a href="{{ pathto('license') }}">License</a></li>
</ul>

<h3>Cortix Links</h3>
<ul>
  <li><a href="https://mybinder.org/v2/gh/dpploy/cortix-nb/master">Run examples: Binder</a></li>
  <li><a href="https://nbviewer.jupyter.org/github/dpploy/cortix-nb/tree/master/">View examples: NBViewer</a></li>
  <li><a href="https://github.com/dpploy/cortix-nb">Examples repo: GitHub</a></li>
  <li><a href="https://pypi.org/project/cortix/">PyPI install</a></li>
  <li><a href="https://github.com/dpploy/cortix">GitHub repo</a></li>
  <li><a href="https://github.com/dpploy/cortix/issues">Issue Tracker</a></li>
</ul>

<h3>HPC Resources</h3>
<ul>
  <li><a href="https://hpc.inl.gov/SitePages/Home.aspx">Idaho National Laboratory HPC</a></li>
  <li><a href="http://wiki.umassrc.org/wiki/index.php/Main_Page">UMass Green HPC Cluster </a></li>
</ul>
